"""Pyxis.ModSupport: functions for programming Pyxides modules"""

import fnmatch
import inspect, traceback

import Pyxis
from Pyxis import *

from Pyxis.Commands import _verbose,_warn,_abort,makedir
from Pyxis.Internals import _superglobals,_namespaces,_modules

  
def register_pyxis_module (superglobals=""):
  """Registers a module (the callee) as part of the Pyxis environment.
  'superglobals' can be a list of super-global variables defined by the module.
  Superglobals are propagated across all modules that register for them.
  It is also possible to generate them as v.define() instead.""";
  frame = inspect.currentframe().f_back;
  globs = frame.f_globals;
  modname = globs['__name__'];
  module = sys.modules[modname];
  if modname.startswith("Pyxides."):
    modname = modname.split(".",1)[-1];
  if id(globs) in _superglobals:
    if _modules[modname] is not module:
      raise RuntimeError("a different Pyxis module named '%s' is already registered"%modname);
    return
  _modules[modname] = module;
  # build list of superglobals
  if isinstance(superglobals,str):
    superglobs = superglobals.split();
  else:
    superglobs = itertools.chain(*[ x.split() for x in superglobals ]);
  superglobs = set(superglobs);
  _verbose(1,"registered module '%s'"%modname);
  _namespaces[modname] = globs;
  _superglobals[id(globs)] = superglobs;
  Pyxis.Context[modname] = module;
  # add superglobals
  for sym in superglobs:
    # if superglobal is already defined, copy its value to the new module
    # if superglobal was not yet defined, get its value form the module (or use None),
    # and propagate this value super-globally via assign
    if sym in Pyxis.Context:
      globs[sym] = Pyxis.Context[sym];
    else:
      assign(sym,globs.get(sym,None),namespace=Pyxis.Context,frame=frame)
  # report 
  Pyxis.Internals.report_symbols(modname,superglobs,
      [ (name,obj) for name,obj in globs.items() if not name.startswith("_") and name not in Pyxis.Commands.__dict__ and name not in superglobs ]);
      

def def_global (name,default,doc=None,config=False):
  """Defines a module global with the given name, default value and documentation string.
  Mainly useful in Pyxides modules, to provide documentation on their globals.
  At module level, calling def_global('NAME',value,doc) is equivalent to
      NAME = value
      _doc_NAME = doc
  Also, if config=True, then the global variable refers to a config file that will be automatically
  backed up to OUTDIR.
  """;
  globs = inspect.currentframe().f_back.f_globals;
  globs[name] = default;
  if name.endswith("_Template"):
    name = name[:-9];
  globs.setdefault("_symdocs",{})[name] = doc;
  if config:
    globs.setdefault("_config_files",[]).append(name);
  
define = def_global;


import itertools
  
def document_globals (obj,*patterns):
  """Updates an object's documentation string with a list of globals that match the specified patterns.
  Can take multiple pattern arguments; each argument will be split at whitespace.
  Mainly useful in Pyxides modules, to form documentation strings for functions, e.g.:
  
    def_global(FOO_A,"1","option A for function foo");
    def_global(FOO_B,"2","option B for function foo");
    
    def foo ():
      # uses FOO_A and FOO_B
    
    document_globals(foo,"FOO_*");
    
  This results in foo.__doc__ being extended with documentation for FOO_A and FOO_B
  """
  patterns = list(itertools.chain(*[ patt.split() for patt in patterns ]));
  globs0 = inspect.currentframe().f_back.f_globals;
  modname0 = globs0['__name__'];
  globdocs = Pyxis.Context.get('_symdocs',{});
  # for each module, make a set of all global symbols, plus symbols that may not yet be defined, but already have documentation
  # (i.e. have entries in _symdocs)
  # each entry in allsyms is a pair of set_of_symbols,dict_of_docs
  allsyms = {}
  for modname,globs in _namespaces.items():
    docs = globs.get('_symdocs',{});
    allsyms[modname] = set(itertools.chain(iter(globs.keys()),iter(docs.keys()))),docs;
  if modname0 not in allsyms:
    docs = globs0.get('_symdocs',{});
    allsyms[modname0] = set(itertools.chain(iter(globs0.keys()),iter(docs.keys()))),docs;
  # keep track of what's been documented, to avoid duplicates
  documented = set();
  doclist = [];
  for patt in patterns:
    # if patt matches a superglobal defined in calling module, add to list as a superglobal
    if Pyxis.Internals.is_superglobal(globs0,patt):
      doclist += [ ("v."+patt,globdocs.get(patt,'')) ];
    # else look for matching module-level globals
    else:
      if '.' in patt:
        modname,patt = patt.split('.',1);
        if modname not in allsyms:
          raise ValueError("document_globals(\"%s.%s\"): module '%s' not registered"%(modname,patt,modname));
      else:
        modname = modname0;
      modsyms,moddocs = allsyms[modname];
      # make sorted list of matching module globals for the pattern, and loop over them
      for sym in sorted([ sym for sym in modsyms if fnmatch.fnmatch(sym,patt) and not sym.endswith("_Template") ]):
        fqsym = "%s.%s"%(modname,sym);
        # add to doclist, unless already documented
        if fqsym not in documented:
          doclist.append((fqsym,moddocs.get(sym,'')));
        documented.add(fqsym);
  # now generate documentation string
  if doclist:
    text = "\nThe following variables also apply:\n\n"
    for sym,doc in doclist:
      text += "  %-20s %s\n"%(sym,doc);
    obj.__doc__ = (obj.__doc__ or "")+ text;
  
def interpolate_locals (*varnames):
  """interpolates the variable names (from the local context) given by its argument(s).
  Returns new values in the order given. Useful as the opening line of a function, for example:
  
  def f(a="$A",b="$a.1"):
    a,b = interpolate_locals("a b")   # can also be specified as "a","b"
    
  will assign the global variable A to a, and the value of a (in this case, the value of A) plus ".1" to b.
  """;
  ## NB: the rationale for implementing it like this, as opposed to directly manipulating f_locals
  ## of the caller frame, is because f_locals of the caller can be read-only depending on Python implementation.
  Pyxis.Internals.assign_templates();
  # interpolate the whole locals() dict
  frame = inspect.currentframe().f_back;
  locs = interpolate(frame.f_locals,frame,depth=2);
  # return variables in the order listed
  ret = [ locs.get(name) for name in itertools.chain(*[ v.split(" ") for v in varnames ]) ];
  return ret if len(ret) != 1 else ret[0];
    
def kwopt_to_command_line (**kwopt):
  """Converts keyword options to command-line arguments: returns a string of 
  --NAME=VALUE or --NAME (for value=True) pairs, where NAME and VALUE are looked up in the dict, and
  VALUE is interpolated."""
  return " ".join([ ("--%s %s"%(name,interpolate(value,depth=2)) if value is not True else "--%s"%name) for name,value in kwopt.items()]);


from Pyxis.ModSupport import *

# register ourselves with Pyxis and define the superglobals
#register_pyxis_module(superglobals="MS LSM OUTDIR DESTDIR")

register_pyxis_module('im')

# external tools  
define('IMAGER','lwimager','imager to use. Default is lwimager.')
define('LWIMAGER_PATH','lwimager','path to lwimager binary. Default is to look in the system PATH.')
define('WSCLEAN_PATH','wsclean','path to wsclean binary.')
define('MORESANE_PATH','runsane','path to PyMORESANE')
define('CASA_PATH','casapy','path to casapy')

# default clean algorithm
define("CLEAN_ALGORITHM","clark","CLEAN algorithm (clark, hogbom, csclean, etc.)")
define('DECONV_LABEL',None,'Label to identify images from different deconvolution algorithms. If set to True, name of algorithm will be used.')
# Default imaging colun
define('COLUMN','CORRECTED_DATA','default column to image')

# filenames for images
define("BASENAME_IMAGE_Template","${OUTFILE}${-<IMAGER}","default base name for all image filenames below")
define("DIRTY_IMAGE_Template", "${BASENAME_IMAGE}.dirty.fits","output filename for dirty image")
define("PSF_IMAGE_Template", "${BASENAME_IMAGE}.psf.fits","output filename for psf image")
define("RESTORED_IMAGE_Template", "${BASENAME_IMAGE}${-<DECONV_LABEL}.restored.fits","output filename for restored image")
define("RESIDUAL_IMAGE_Template", "${BASENAME_IMAGE}${-<DECONV_LABEL}.residual.fits","output filename for deconvolution residuals")
define("MODEL_IMAGE_Template", "${BASENAME_IMAGE}${-<DECONV_LABEL}.model.fits","output filename for deconvolution model")
define("FULLREST_IMAGE_Template", "${BASENAME_IMAGE}${-<DECONV_LABEL}.fullrest.fits","output filename for LSM-restored image")
define("MASK_IMAGE_Template", "${BASENAME_IMAGE}.mask.fits","output filename for CLEAN mask")

# How to channelize the output image. 0 for average all, 1 to include all, 2 to average with a step of 2, etc.
# None means defer to 'imager' module options
define("IMAGE_CHANNELIZE",0,"image channels selection: 0 for all, 1 for per-channel cube")
# passed to tigger-restore when restoring models into images. Use e.g. "-b 45" for a 45" restoring beam.
define("RESTORING_OPTIONS","","extra options to tigger-restore for LSM-restoring")

# Standard imaging options
DOUBLE_PSF = True
ifrs = ""
npix = 1024
cellsize = "2arcsec"
mode = "channel"
stokes = "IQUV"
weight = "briggs"
robust = 0
niter = 1000
gain = .1
threshold = 0
wprojplanes = 0
cachesize = 4096
fixed = 0
# rescale images by factor
flux_rescale=1
# use velocity rather than frequency
velocity = False 
no_weight_fov = False

# import wsclean
# import lwimager
# import casa
# import moresane

def make_image (*args,**kw):
  imager = kw.get('imager', IMAGER).lower();
  imgmod = "im."+imager;
  try:
    mod = getattr(__import__(imgmod),imager);
  except:
    traceback.print_exc();
    abort("import $imgmod failed")
  call_imager = getattr(mod,'make_image',None);
  if not callable(call_imager):
    abort("$imgmod does not provide a make_image() function");
  call_imager(*args,**kw);






import sys
import traceback
import inspect
import os
import os.path
import time
import math
import multiprocessing
import queue

import Pyxis
import Pyxis.Internals
from Pyxis.Internals import _int_or_str,interpolate,assign


DEG = math.pi/180
ARCMIN = DEG/60
ARCSEC = ARCMIN/60

def _init (context):
  global x,xo,xz,xr,xro,v,E;
  # add some standard objects
  # the 'x' object is a shortcut for executing shell commands. E.g. x.ls('.')
  # the 'xo' object is a shortcut for executing shell commands that are allowed to fail. E.g. x.ls('.')
  x = Pyxis.Internals.ShellExecutorFactory(allow_fail=False);
  x.__name__ = 'x';
  x.__doc__ = x.doc_proto%dict(name='x') + """Shell commands launched via 'x' must succeed for the script 
to continue. Upon error, the current script is aborted.""";

  xo = Pyxis.Internals.ShellExecutorFactory(allow_fail=True);
  xo.__name__ = 'xo';
  xo.__doc__ = xo.doc_proto%dict(name='xo') + """Shell commands launched via 'xo' are allowed to fail, with
the script continuing regardless.""";

  xr = Pyxis.Internals.ShellExecutorFactory(get_output=True,verbose=3);
  xr.__name__ = 'xr';
  xr.__doc__ = xr.doc_proto%dict(name='xr') + """Shell commands launched via 'xr' return their output as
a Python string.""";

  xro = Pyxis.Internals.ShellExecutorFactory(get_output=True,allow_fail=True,verbose=0);
  xro.__name__ = 'xr';
  xro.__doc__ = xro.doc_proto%dict(name='xro') + """Shell commands launched via 'xro' return their output as
a Python string, and are allowed to fail.""";

  xz = Pyxis.Internals.ShellExecutorFactory(allow_fail=True,bg=True);
  xz.__name__ = 'xz';
  xz.__doc__ = xz.doc_proto%dict(name='xz') + """Shell commands launched via 'xz' are run in the background,
in parallel with the rest of the script. They are allowed to fail, with the script continuing regardless.""";

  v = Pyxis.Internals.GlobalVariableSpace(context);
  object.__setattr__(v,'__name__','v');
  
  E = Pyxis.Internals.ShellVariableSpace();
  object.__setattr__(E,'__name__','E');


def _I (string,level=1):
  """_I(string): interpolates (i.e. replaces by value) $VAR and ${VAR} occurrences of local and 
  global variables. Local variables are interpreted as those in the context of the caller (i.e. 1 level away).
  
  _i(string,level): change the number of caller levels for lookup of locals, e.g. 2 means caller of caller
  _i(string,-1): interpolate across all callers
  """;
  frame = inspect.currentframe();
  for i in range(level):
    frame = frame and frame.f_back;
  return interpolate(string,frame);
  
def _II (*strings):
  """_II(string): interpolates multiple strings. Does an implicit assign_templates call beforehand. Useful as
  the opening line of a function, as e.g. arg1,arg2 = _II(arg1,arg2);
  """;
  Pyxis.Internals.assign_templates();
  frame = inspect.currentframe().f_back;
  ret = [ x and interpolate(x,frame) for x in strings ];
  return ret[0] if len(strings)<2 else ret;

II = _II;  
  
_subprocess_id = None;  
  
def _timestamp ():
  ts = time.strftime("%Y/%m/%d %H:%M:%S");
  if _subprocess_id is not None:
    ts += " [%d]"%_subprocess_id;
  return ts;

def _message (*msg,**kw):
  output = " ".join(map(str,msg));
  sync = kw.get("sync");
  quiet = ( Pyxis.Context.get('QUIET') or kw.get("quiet") ) and not kw.get('critical'); 
  if sys.stdout is not sys.__stdout__:
    if sync:
      fcntl.lockf(sys.stdout,fcntl.LOCK_EX);
      sys.stdout.seek(0,2);
    print(output);
    if sync:
      sys.stdout.flush();
      fcntl.lockf(sys.stdout,fcntl.LOCK_UN);
    if not quiet:
      sys.__stdout__.write(output+"\n");
  else:
    if not quiet:
      print(output);
  
def _debug (*msg,**kw):
  """Prints debug message(s) without interpolation""";
  _message(_timestamp(),"DEBUG:",*msg,**kw);

def _verbose (level,*msg,**kw):
  """Prints verbosity message(s), if verbosity level is >= Context.pyxis_verbosity""";
  try:
    verb = int(Pyxis.Context['VERBOSE']);
  except:
    verb = 1;
  try:
    level = int(level)
  except:
    level = 1;
  if level <= verb:
    _message(_timestamp(),"PYXIS:",*msg,**kw);

def _info (*msg,**kw):
  """Prints info message(s) without interpolation""";
  _message(_timestamp(),"INFO:",*msg,**kw);

def _warn (*msg,**kw):
  """Prints warning message(s) without interpolation""";
  _message(_timestamp(),"WARNING:",*msg,**kw);

def _error (*msg,**kw):
  """Prints error message(s) without interpolation""";
  _message(_timestamp(),"ERROR:",*msg,**kw);

def _abort (*msg,**kw):
  """Prints error message(s) without interpolation and aborts""";
  _message(_timestamp(),"ABORT:",console=True,critical=True,*msg,**kw);
  Pyxis.Internals.flush_log();
  sys.exit(1);
  
def debug (*msg):
  """Prints debug message(s)""";
  _debug(*[ _I(x,2) for x in msg ]);

def verbose (level,*msg):
  """Prints verbosity message(s), if verbosity level is >= Context.pyxis_verbosity""";
  _verbose(level,*[ _I(x,2) for x in msg ]);

def info (*msg):
  """Prints info message(s)""";
  _info(*[ _I(x,2) for x in msg ]);

def warn (*msg):
  """Prints warning message(s)""";
  _warn(*[ _I(x,2) for x in msg ]);

def error (*msg):
  """Prints error message(s)""";
  _error(*[ _I(x,2) for x in msg ]);

def abort (*msg):
  """Prints error message(s) and aborts""";
  _abort(*[ _I(x,2) for x in msg ]);
  
def pyxreload ():
  from Pyxis.Internals import _modules
  for m in 'Pyxis.Internals','Pyxis.Commands','Pyxis.ModSupport':
    info("Reloading",m);
    reload(sys.modules[m]);
  for m in _modules.values():
    info("Reloading",m.__name__);
    reload(m);

def pyxlf (mod=None):
  """Prints all functions defined by module""";
  pyxls(mod,what="F");

def pyxlv (mod=None):
  """Prints all global variables defined by module""";
  pyxls(mod,what="V");

def pyxls (mod=None,what="FVTB"):
  import Pyxis.ModSupport
  """Prints symbols defined by module. 'what' is a combination of characters specifying what to print.""";
  # make sorted list of globals (excepting those starting with underscore)
  globs = Pyxis.Context if mod is None else vars(mod);
  globs = [ (name,value) for name,value in sorted(globs.items()) 
            if name[0]!="_" and name not in Pyxis._predefined_names ];
  # from this, extract the non-callables
  varlist = [ (name,value) for name,value in globs if not callable(value)  ];
  if "V" in what:
    print("Globals:");
    for var,val in varlist:
      if not var.endswith("_List") and not var.endswith("_Template") and isinstance(val,(str,int)):
        print("  %s=%s"%(var,val));
    print("Lists:");
    for var,val in varlist:
      if var.endswith("_List"):
        print("  %s=%s"%(var,val));
    print("Templates:");
    for var,val in globs:
      if var.endswith("_Template"):
        print("  %s=%s"%(var,val));
  # now deal with the callables
  if "F" in what:
    print("Functions:");
    print(" ",", ".join([ name for name,value in globs 
        if callable(value) and not isinstance(value,Pyxis.Internals.ShellExecutor) and not name.endswith("_Template") 
          and not name in globals() 
          and not name in Pyxis.ModSupport.__dict__
      ]));
  if "T" in what:
    print("External tools:");
    for name,value in globs:
      if isinstance(value,Pyxis.Internals.ShellExecutor):
        print("  %s=%s"%(name,value.path));
  if "B" in what:
    print("Pyxis built-ins:");
    print(" ",", ".join([ name for name,value in globs 
      if callable(value) and not isinstance(value,Pyxis.Internals.ShellExecutor) and not name.endswith("_Template") and name in globals() ]));

def exists (filename):
  """Returns True if filename exists, interpolating the filename""";
  return filename and os.path.exists(_I(filename,2));

def _per (varname,parallel,*commands):
  # default frame to look for vars is caller of caller
  frame = inspect.currentframe().f_back.f_back;
  namespace,vname = Pyxis.Internals._resolve_namespace(varname,frame=frame,default_namespace=Pyxis.Context);
  _verbose(2,"per(%s.%s)",namespace.get('__name__',"???") if namespace is not Pyxis.Context else "v",vname);
  saveval = namespace.get(vname,None);
  def _restore ():
#    if saveval is not None:
    _verbose(2,"restoring %s=%s"%(varname,saveval),sync=True);
    assign(varname,saveval,namespace=namespace,interpolate=False);
  varlist = namespace.get(vname+"_List",None);
  cmdlist = ",".join([ x if isinstance(x,str) else getattr(x,"__name__","?") for x in commands ]);
  persist = Pyxis.Context.get("PERSIST");
  fail_list = [];
  if varlist is None:
    _verbose(1,"per(%s,%s): %s_List is empty"%(varname,cmdlist,varname));
    return;
  try:
    if type(varlist) is str:
      varlist = list(map(_int_or_str,varlist.split(",")));
    elif not isinstance(varlist,(list,tuple)):
      _abort("PYXIS: per(%s,%s): %s_List has invalid type %s"%(varname,cmdlist,str(type(varlist))));
    nforks = Pyxis.Context.get("JOBS",0);
    stagger = Pyxis.Context.get("JOB_STAGGER",0);
    # unforked case
    _verbose(1,"per(%s,%s,persist=%d): iterating over %s=%s"%(varname,cmdlist,1 if persist else 0,varname," ".join(map(str,varlist))));
    global _subprocess_id;
    if not parallel or nforks < 2 or len(varlist) < 2 or _subprocess_id is not None:
      # do the actual iteration
      for value in varlist:
        _verbose(1,"per-loop, setting %s=%s"%(varname,value));
        assign(vname,value,namespace=namespace,interpolate=False); 
        try:
          Pyxis.Internals.run(*commands);
        except (Exception,SystemExit,KeyboardInterrupt) as exc:
          if persist:
            _warn("exception raised for %s=%s:\n"%(vname,value),
                *traceback.format_exception(*sys.exc_info()));
            _warn("persistent mode is on (PERSIST=1), so continuing to end of %s_List"%vname)
            fail_list.append((value,str(exc)));
          else:
            raise;
      # any fails?
      if fail_list:
        _restore();
        _abort("per-loop failed for %s"%(",".join([f[0] for f in fail_list])));
    else:
      # else split varlist into forked subprocesses
      nforks = min(nforks,len(varlist));
      # create a queue for all variable values
      varqueue = multiprocessing.Queue(len(varlist))
      for x in varlist:
        varqueue.put(x)
      # distribute N values per each fork
      _verbose(1,"splitting into %d jobs by %s, staggered by %ds"%(nforks,varname,stagger));
      Pyxis.Internals.flush_log();
      forked_pids = {};
      try:
        for job_id in range(nforks):
          if job_id and stagger:
            time.sleep(stagger);
          # subvals is range of values to be iterated over by this subjob
          pid = os.fork();
          if not pid:
            # child fork: run commands while something is on queue
            _subprocess_id = job_id;
            _verbose(1,"started job %d"%job_id,sync=True);
            try:
              fail_list = []
              success_list = []
              while True:
                try:
                  value = varqueue.get(False)
                except queue.Empty:
                  break
                _verbose(1,"per-loop, setting %s=%s"%(varname,value),sync=True);
                assign(vname,value,namespace=namespace,interpolate=False);
                try:
                  Pyxis.Internals.run(*commands);
                  success_list.append(value)
                except (Exception,SystemExit,KeyboardInterrupt) as exc:
                  if persist:
                    _warn("exception raised for %s=%s:\n"%(vname,value),
                        sync=True,*traceback.format_exception(*sys.exc_info()));
                    _warn("persistent mode is on (PERSIST=1), so continuing to end of %s_List"%vname,sync=True)
                    fail_list.append((value,str(exc)));
                  else:
                    raise;
              # any successes?
              if success_list:
                _verbose(1,"job #%d (pid %d): per-loop succeeded for %s"%(_subprocess_id,pid,
                    ", ".join([str(f) for f in success_list])),sync=True)
              # any fails?
              if fail_list:
                _restore();
                _abort("job #%d (pid %d): per-loop failed for %s"%(_subprocess_id,pid,
                    ", ".join([str(f[0]) for f in fail_list])),sync=True)
            except:
              traceback.print_exc();
              _verbose(1,"job #%d (pid %d) aborted at %s=%s, exiting with error code 1"%(_subprocess_id,pid,varname,value),sync=True);
              _restore();
              _verbose(2,"logfile is",Pyxis.Context.get('LOG'),sync=True);
              _error("per-loop failed for %s"%value,sync=True);
              sys.exit(1);
            _verbose(2,"job #%d (pid %d) exiting normally"%(_subprocess_id,os.getpid()),sync=True);
            sys.exit(0);
          else: # parent pid: append to list
            _verbose(2,"launched job #%d with pid %d"%(job_id,pid),sync=True);
            forked_pids[pid] = job_id
        njobs = len(forked_pids);
        _verbose(1,"%d jobs launched, waiting for finish"%len(forked_pids),sync=True);
        failed = [];
        while forked_pids:
          pid,status = os.waitpid(-1,0);
          if pid in forked_pids:
            job_id = forked_pids.pop(pid);
            status >>= 8;
            if status:
              failed.append(job_id);
  #            success = False;
              _error("job #%d exited with error status %d, waiting for %d more jobs to complete"%(job_id,status,len(forked_pids)),sync=True);
            else:
              _verbose(1,"job #%d finished, waiting for %d more jobs to complete"%(job_id,len(forked_pids)),sync=True);
        if failed:
          _abort("%d of %d jobs have failed"%(len(failed),njobs),sync=True);
        else:     
          _verbose(1,"all jobs finished ok",sync=True);
      except KeyboardInterrupt:
        if _subprocess_id is None:
          _restore();
          _error("Caught Ctrl+C, waiting for %d jobs to exit"%len(forked_pids),sync=True);
          import signal;
          for pid in list(forked_pids.keys()):
            os.kill(pid,signal.SIGINT);
          while forked_pids:
            pid,status = os.waitpid(-1,0);
            if pid in forked_pids:
              job_id = forked_pids.pop(pid);
              _verbose(1,"job #%d exited with error status %d, waiting for %d more"%
                  (job_id,status>>8,len(forked_pids)),sync=True);
        raise;
  finally:
    # note that children also execute this block with sys.exit()
    if _subprocess_id is None:
      _restore();
    Pyxis.Internals.flush_log();

def per (varname,*commands):
  """Iterates over variable 'varname', and executes commands. That is, for every value
  in varname_List, sets varname to that value, then calls the commands.
  Uses non-parallel mode."""
  _per(varname,False,*commands);

def pper (varname,*commands):
  """Iterates over variable 'varname', and executes commands. That is, for every value
  in varname_List, sets varname to that value, then calls the commands.
  Uses parallel mode."""
  _per(varname,True,*commands);

def per_ms (*commands):
  """Iterates over variable 'MS', and executes commands. That is, for every value
  in MS_List, sets MS to that value, then calls the commands. Default mode is parallel."""
  _per("MS",True,*commands);

def per_ddid (*commands):
  """Iterates over variable 'DDID', and executes commands. That is, for every value
  in DDID_List, sets MS to that value, then calls the commands. Default mode is serial"""
  _per("ms.DDID",False,*commands);

def is_true (arg):
  """Returns True if argument evaluates to boolean truth, possibly as a string.
  is_true('False') == is_true('0') == is_true(0) == False
  is_true('True') == is_true('1') == is_true(1) == True
  """;
  arg = _I(arg,2);
  if isinstance(arg,str):
    if not arg:
      return False;
    try:
      return bool(eval(arg));
    except:
      raise TypeError("is_true('%s'): invalid string argument"%arg);
  try:
    return bool(arg);
  except:
    raise TypeError("is_true(%s): invalid argument %s"%(str(arg),str(type(arg))));

def makedir (dirname,no_interpolate=False):
  """Makes sure the supplied directory exists, by creating parents as necessary. Interpolates the dirname.""";
  if not no_interpolate:
    dirname = interpolate(dirname,inspect.currentframe().f_back);
  parent = dirname;
  while parent and parent[-1] == '/':
    parent = parent[:-1]
  # go back and accumulate list of dirs to be created
  parents = [];
  while parent and not os.path.exists(parent):
    parents.append(parent);
    parent = os.path.dirname(parent);
  # create in reverse
  for parent in parents[::-1]:
    verbose(1,"creating directory %s"%parent);
    os.mkdir(parent);
    
    
import tempfile   
import pickle
import fcntl
    
class Safelist (object):
  """A Safelist provides a list of objects that is shared among parallel processes. Multiple jobs launched
  by pyxis can append to a Safelist in a multiprocess-safe manner: write locks are enforced""";
  def __init__ (self,filename=None):
    """Creates a safelist, associated with the given filename. If not supplied, a filename is
    chosen randomly"""
    if filename:
      filename = _I(filename,2);
      if os.path.exists(filename):
        os.remove(filename);
      self.filename = filename;
    else:
      self.filename = tempfile.NamedTemporaryFile(delete=False).name;

  def reset (self):
    """Resets safelist to empty (by deleting the associated file)""";
    if os.path.exists(self.filename):
      os.remove(self.filename);
      
  def add (self,obj):
    """Adds an object to the safelist, in an MP-safe manner""";
    if isinstance(obj,str):
      obj = _I(obj,2);
    ff = open(self.filename,"ab");
    fcntl.flock(ff,fcntl.LOCK_EX);
    try:
      pickle.dump(obj,ff);
    finally:
      fcntl.flock(ff,fcntl.LOCK_UN);
    
  def read (self):
    """Reads all objects accumulated in the safelist, in an MP-safe manner""";
    ret = [];
    if os.path.exists(self.filename):
      ff = open(self.filename,"rb");
      fcntl.flock(ff,fcntl.LOCK_EX);
      try:
        while True:
          ret.append(pickle.load(ff));
      except EOFError:
        pass;
      finally:
        fcntl.flock(ff,fcntl.LOCK_UN);
    return ret;
    


import inspect
import os.path
import math
import numpy
import imp
np = numpy

# some useful constants
DEG = math.pi/180
ARCMIN = DEG/60
ARCSEC = DEG/3600

#from Pyxis.Commands import *

import Pyxis.Internals 

_initialized = False;

if not _initialized:
  """Initializes Pyxis with a global context dict.""";
  # initialize Pyxis using the globals of whoever imported the module
  ## OMS: this was enough in Python2. Not in Py3
  #_context = inspect.currentframe().f_back.f_globals;
  ## in Py3:
  frame = inspect.currentframe().f_back
  while frame:
    _context = frame.f_globals
    if "PYXIS_ROOT_NAMESPACE" in _context:
      print("root namespace is",_context.get("__name__"),_context.get("__path__"))
      break
    frame = frame.f_back

  if not frame:
    raise RuntimeError("Unable to find root namespace. Set PYXIS_ROOT_NAMESPACE=True before importing pyxis.")

  Pyxis.Internals.init(_context);  
  from Pyxis.Commands import *
 
  verbose(1,"===[ Pyxis: Python eXtensions for Inteferometry Scripting (C) 2013 by Oleg Smirnov <oms@ska.ac.za> ]===");
  if Pyxis.Internals._verbose_startup_message:
    verbose(*Pyxis.Internals._verbose_startup_message);

  # import basic Pyxis commands into the context
  verbose(1,"loading Pyxis into context '%s'"%_context.get('__name__'));
  exec('from Pyxis.Commands import *',_context);
  exec('from Pyxis.Commands import _I,_II',_context);
  
  pyxides_path = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "Pyxides"))
  verbose(2,"importing Pyxides from",pyxides_path)

  Pyxides = imp.load_source('Pyxides',os.path.join(pyxides_path, "__init__.py"))
  # add Pyxides path to module includes (so we can do stuff like "import ms" instead of "from Pyxides import ms"
  if _context.get("ADD_PYXIDES_PATH",True):
    verbose(2,"added %s to import path. Set ADD_PYXIDES_PATH=False to disable"%pyxides_path);
    sys.path.append(pyxides_path)

  ## import standard modules, unless a specific other set is given
  #if not context.get("pyxis_preload"):
    #import StandardModules
    #for mod in StandardModules.pyxis_preload:
      #filename = os.path.join(os.path.dirname(StandardModules.__file__),mod)+".py";
      #if os.path.exists(filename):
        #verbose(1,"loading standard package '%s' from %s"%(mod,filename));
        #Pyxis.Internals.load_package(mod,filename);        
      #else:
        #warn("can't find standard package %s"%filename);
  
  Pyxis.Internals.initconf();
  



from Pyxis.ModSupport import *

import ms
import mqt
import imager
import std

from astropy.io import fits as pyfits

# register ourselves with Pyxis, and define what superglobals we use (these come from ms)
register_pyxis_module(superglobals="MS LSM DESTDIR OUTFILE");
  


#!/usr/bin/env python

import os
from distutils.core import setup
import six

install_requires = [
      'astropy<=2.0.11',
      # 'timba', not available on pypi
      'matplotlib',
      'python_casacore',
      'numpy<=1.16',
      'scipy',
      'future'
] if six.PY2 else [
      'astropy>=3.0',
      # 'timba', not available on pypi
      'matplotlib',
      'python_casacore',
      'numpy>=1.16',
      'scipy',
      'future'
]

def readme():
    """ Return README.rst contents """
    with open('README.md') as f:
      return f.read()

setup(name='astro-pyxis',
      version='1.7.3',
      python_requires='>=3.0.0',
      description='Python Extensions for astronomical Interferometry Scripting',
      author='Oleg Smirnov',
      author_email='Oleg Smirnov <osmirnov@gmail.com>',
      url='https://github.com/ska-sa/pyxis',
      packages=['Pyxis', 'Pyxides', 'Pyxides.im', 'Pyxides.utils'],
      install_requires=install_requires,
      zip_safe=True,
      include_package_data=True,
      scripts=['Pyxis/bin/' + i for i in os.listdir('Pyxis/bin')],
     )
